/////////////////////////////////////////////////////////////////////////////////////////
// Jocopod -- Joystick to Control Port Dongle
//            --          --      --   -
//
// (c) 2015-2016, Oliver "lvr" Tscherwitschke
// lvr@bitbang.de
//
// This file is part of the Jocopod Firmware.
//
// The Jocopod Firmware is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// The Jocopod Firmware is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with the Jocopod Firmware. If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////
// Include files
#include <EEPROM.h>
#include <limits.h>

#include "mytypes.h"
#include "config.h"
#include "main.h"
#include "io.h"
#include "eedata.h"
#include "util.h"


/////////////////////////////////////////////////////////////////////////////////////////
// Variables

uint8_t operatingMode = MODE_PADDLE;    
                                                
Axis axisX = {
    .joy = {.min = 5, .mid = 300, .max = 500},
    .pot = {.min = 0, .mid = 127, .max = 255},
    .invert = true
};

Axis axisY = {
    .joy = {.min = 5, .mid = 300, .max = 500},
    .pot = {.min = 0, .mid = 127, .max = 255},
    .invert = false
};

enum State {NORMAL, CALIB_ENTER, CALIB, CENTER, OCR_OUT};
static State state = NORMAL;

volatile int ocrX;
volatile int ocrY;

int led1State = LOW;
int led2State = LOW;


// -------------------------------------
// median filter
int medianBufX[MEDIAN_BUFLEN];
int medianBufY[MEDIAN_BUFLEN];
unsigned short medianIdx;
int joy1X_median;
int joy1Y_median;
// -------------------------------------


// -------------------------------------
// moving average filter
int bufx[MA_BUFLEN];
int bufy[MA_BUFLEN];
unsigned short idx;
#if MA_BUFLEN > (INT_MAX / MAX_VALID_JOY_VALUE)
long sumx, sumy;
#else
int sumx, sumy;
#endif
int joy1X_mean;
int joy1Y_mean;
// -------------------------------------



// Threshoolds for hysteresis in digital modes
int thresholdsLeft[3];
int thresholdsRight[3];
int thresholdsUp[3];
int thresholdsDown[3];


/////////////////////////////////////////////////////////////////////////////////////////
// Functions

// Compares two ints, needed for qsort().
int compfunc(const void* a, const void* b)
{
    if (*(int*) a < *(int*) b)
        return -1;
    else if (*(int*) a > *(int*) b)
        return 1;
    else
        return 0;
}


// Calculates the median of buf[]. Buf has a length of MEDIAN_BUFLEN.
int median(int buf[])
{
    static int tmpbuf[MEDIAN_BUFLEN];

    memcpy(tmpbuf, buf, sizeof tmpbuf);
    qsort(tmpbuf, MEDIAN_BUFLEN, sizeof tmpbuf[0], compfunc);
    
    return tmpbuf[MEDIAN_BUFLEN / 2];
}


// Linearize the joystick position to AD value curve
int linearize(int ad)
{
    long res = (1023L - ad) * 1024 / (ad + 1);
    
    return min(res, INT_MAX);
}


// Init thresholds for hysteresis             
void initThresholds(void)
{
    int mi;

    // Exit if we are not in a digital mode
    if (operatingMode != MODE_JOY_WHEEL && operatingMode != MODE_JOY_STICK)
        return;

    mi = operatingMode - MODE_JOY_WHEEL;
    
    thresholdsLeft[MID]  = axisX.joy.mid - (axisX.joy.mid - axisX.joy.min) * thDef[mi].xThresh;
    thresholdsRight[MID] = axisX.joy.mid + (axisX.joy.max - axisX.joy.mid) * thDef[mi].xThresh;
    thresholdsUp[MID]    = axisY.joy.mid - (axisY.joy.mid - axisY.joy.min) * thDef[mi].yThresh;
    thresholdsDown[MID]  = axisY.joy.mid + (axisY.joy.max - axisY.joy.mid) * thDef[mi].yThresh;

    thresholdsLeft[NEG]  = axisX.joy.mid - (axisX.joy.mid - axisX.joy.min) * (thDef[mi].xThresh - thDef[mi].xHyst);
    thresholdsRight[NEG] = axisX.joy.mid + (axisX.joy.max - axisX.joy.mid) * thDef[mi].xThresh;
    thresholdsUp[NEG]    = axisY.joy.mid - (axisY.joy.mid - axisY.joy.min) * (thDef[mi].yThresh - thDef[mi].yHyst);
    thresholdsDown[NEG]  = axisY.joy.mid + (axisY.joy.max - axisY.joy.mid) * thDef[mi].yThresh;
     
    thresholdsLeft[POS]  = axisX.joy.mid - (axisX.joy.mid - axisX.joy.min) * thDef[mi].xThresh;
    thresholdsRight[POS] = axisX.joy.mid + (axisX.joy.max - axisX.joy.mid) * (thDef[mi].xThresh - thDef[mi].xHyst);
    thresholdsUp[POS]    = axisY.joy.mid - (axisY.joy.mid - axisY.joy.min) * thDef[mi].yThresh;
    thresholdsDown[POS]  = axisY.joy.mid + (axisY.joy.max - axisY.joy.mid) * (thDef[mi].yThresh - thDef[mi].yHyst);

#if DEBUG_ENABLE
    Serial.println("\nThresholds:");
    Serial.print("X: left     right\r\n   ");
    Serial.print(thresholdsLeft[MID]);
    Serial.print("        ");
    Serial.println(thresholdsRight[MID]);

    Serial.print("     ");
    Serial.print(thresholdsLeft[NEG]);
    Serial.print("    ");
    Serial.println(thresholdsRight[POS]);


    Serial.print("\nY: up        down\r\n   ");
    Serial.print(thresholdsUp[MID]);
    Serial.print("        ");
    Serial.println(thresholdsDown[MID]);

    Serial.print("     ");
    Serial.print(thresholdsUp[NEG]);
    Serial.print("    ");
    Serial.println(thresholdsDown[POS]);
#endif                
}


// --------------------------------------------------------
// This is called once on boot-up to initialize the system
void setup() 
{
    int rc;
    bool joyBtn1, joyBtn2;
    bool calBtn2 = false;
    
    // Init LED pins
    pinMode(DPIN_LED1, OUTPUT);
    LED1_OFF();

    // Init joystick button input pins
    pinMode(DPIN_JOYBTN1, INPUT_PULLUP);
    pinMode(DPIN_JOYBTN2, INPUT_PULLUP);
    pinMode(DPIN_JOYBTN3, INPUT_PULLUP);
    pinMode(DPIN_JOYBTN4, INPUT_PULLUP);

    // Init digital output pins (high-z)
    pinMode(DPIN_JOY_UP, INPUT);
    pinMode(DPIN_JOY_DOWN, INPUT);
    pinMode(DPIN_JOY_LEFT, INPUT);
    pinMode(DPIN_JOY_RIGHT, INPUT);
    pinMode(DPIN_JOY_FIRE, INPUT);


    // Init ADC
    analogReference(DEFAULT);       // Use Vcc as reference


#if DEBUG_ENABLE
    Serial.begin(115200);
    Serial.println("Jocopod Firmware v" FW_VERSION);
#else
    pinMode(DPIN_LED2, OUTPUT);
    LED2_OFF();

    pinMode(DPIN_CALIB2, INPUT_PULLUP);    
#endif

    // --------------------------------------------------------
    // Read EEPROM
    rc = eepromRead();
#if DEBUG_ENABLE
    Serial.print("eepromRead(): ");
    Serial.println(rc);
#endif


    if (printOpMode(operatingMode))
    {
#if DEBUG_ENABLE
        Serial.println("=> Setting Mode to Paddle.");
#endif        
        operatingMode = MODE_PADDLE;
    }

    printCalib();



    // --------------------------------------------------------
    // Handle switching of operating mode
        
    // If a button is pressed, switch to next operating mode    
    
    joyBtn1 = !digitalRead(DPIN_JOYBTN1);
    joyBtn2 = !digitalRead(DPIN_JOYBTN2) ;
#if !DEBUG_ENABLE
    calBtn2 = !digitalRead(DPIN_CALIB2);
#endif    

    if (joyBtn1 || joyBtn2 || calBtn2)
    {
        if (joyBtn1 || calBtn2) // count mode up
        {
            if (++operatingMode > LAST_VALID_MODE)
                operatingMode = MODE_NULL + 1;
        }
        else if (joyBtn2)       // count mode down
        {
            if (--operatingMode == MODE_NULL)
                operatingMode = LAST_VALID_MODE;
        }

        Serial.print("New ");
        printOpMode(operatingMode);
        eepromWrite();        
    }


    // --------------------------------------------------------
    // Handle if center position is near min or max
    if (fixCalibration())
    {
#if DEBUG_ENABLE
        Serial.println("Fixed calibration:");
        printCalib();
#endif                
    }


    // --------------------------------------------------------
    switch (operatingMode)
    {
        case MODE_PADDLE:
        {
            // Use analog comparator for sync detection
            pinMode(DPIN_DISCHARGE_DETECT, INPUT);  // AIN0
            pinMode(DPIN_DISCHARGE_THRESH, INPUT);  // AIN1
        
            ACSR = (1 << ACIE)    // Analog Comparator Interrupt Enable
                 | (2 << ACIS0);  // Comparator Interrupt on Falling Output Edge
            
            DIDR1 = (1 << AIN1D) | (1 << AIN0D);  // AIN1, AIN0 Digital Input Disable
                
            // Init Timer 1 
            TCNT1 = 0;              // Clear timer
            OCR1A = paddleToOCR(127); 
            OCR1B = paddleToOCR(127); 
            TCCR1A = T1_SET; 
            TCCR1B = T1_START;
        
            // Init pulse pins
            pinMode(DPIN_PADDLE_PULSE_X, OUTPUT);
            digitalWrite(DPIN_PADDLE_PULSE_X, LOW);
        
            pinMode(DPIN_PADDLE_PULSE_Y, OUTPUT);
            digitalWrite(DPIN_PADDLE_PULSE_Y, LOW);

            // Configure pins for full scale output (high-z)
            pinMode(DPIN_PADDLE_FULL_X, INPUT);
            pinMode(DPIN_PADDLE_FULL_Y, INPUT);
            break;
        }

        case MODE_JOY_WHEEL:
        case MODE_JOY_STICK:
        {
            // Init pulse pins
            pinMode(DPIN_PADDLE_PULSE_X, OUTPUT);
            digitalWrite(DPIN_PADDLE_PULSE_X, LOW);
        
            pinMode(DPIN_PADDLE_PULSE_Y, INPUT);

            
            // Configure pins for full scale output (high-z)
            pinMode(DPIN_PADDLE_FULL_X, OUTPUT);
            digitalWrite(DPIN_PADDLE_FULL_X, LOW);

            pinMode(DPIN_PADDLE_FULL_Y, INPUT);

            // Init thresholds for hysteresis             
            initThresholds();
            break;
        }
    }  
}


// --------------------------------------------------------
//
void loop()
{  
    unsigned long now = millis();   // the current time in milli seconds

    // previous joystick position
    static int joy1X_old;
    static int joy1Y_old;

    // current joystick position after linearization [0 .. ~1024] for a 100k pot
    int joy1X;
    int joy1Y;
    int joy2X;
    int joy2Y;

    // joytick buttons
    bool joyBtn1;
    bool joyBtn2;
    bool joyBtn3;
    bool joyBtn4;
    
    // The simulated Paddle value
    int potX;
    int potY;
    static int potX_old;
    static int potY_old;

    bool calBtn1 = false;
    bool calBtn2 = false;
    
    uint16_t tempX;
    uint16_t tempY;

    // init with high value, so that calibration will not trigger, even if calBtn2 is pressed on bootup
    static unsigned long nextTime = 0x40000000; 

/*
 * Some debug stuff. Not guaranteed to do something useful.
 */
#if DEBUG_ENABLE
    static int debug = 0;
    static uint16_t ocrOutX = paddleToOCR(127);
    static uint16_t ocrOutY = paddleToOCR(127);
    int c;
    static int disableFull = 0;

    if (Serial.available() > 0)
    {
        c = Serial.read();

        switch (c)
        {
            case 'v':
                debug = ++debug % 5;
                break;
            
            case 'z':   // Calibrate center position
                state = CENTER;
                break;
            
            case 'c':   // Calib
                if (state == CALIB)
                {
                    state = NORMAL;
                    Serial.println("State: Normal");
                }
                else
                {
                    Serial.println("State: Calib");
                    state = CALIB_ENTER;
                }
                break;

            case 'o':   // direct OCR output
                if (state == OCR_OUT)
                {
                    state = NORMAL;
                    Serial.println("State: Normal");
                }
                else
                {
                    state = OCR_OUT;
                    Serial.println("State: OCR_OUT");
                }
                break;

            case 'q':
                ocrOutX += 100;
                Serial.println(ocrOutX);
                break;

            case 'w':
                ocrOutX += 10;
                Serial.println(ocrOutX);
                break;
            
            case 'e':
                ocrOutX += 1;
                Serial.println(ocrOutX);
                break;
                
            case 'a':
                ocrOutX -= 100;
                Serial.println(ocrOutX);
                break;
                        
            case 's':
                ocrOutX -= 10;
                Serial.println(ocrOutX);
                break;

            case 'd':
                ocrOutX -= 1;
                Serial.println(ocrOutX);
                break;


            case 'j':
                // digitalWrite(DPIN_JOY_LEFT, LOW);
                pinMode(DPIN_JOY_LEFT, OUTPUT);
                delay(500);                
                pinMode(DPIN_JOY_LEFT, INPUT);
                break;

            case 'l':
                pinMode(DPIN_JOY_RIGHT, OUTPUT);
                delay(500);                
                pinMode(DPIN_JOY_RIGHT, INPUT);                 
                break;
            
            case 'i':
                pinMode(DPIN_JOY_UP, OUTPUT);
                delay(500);                
                pinMode(DPIN_JOY_UP, INPUT);                 
                break;

            case 'k':
                pinMode(DPIN_JOY_DOWN, OUTPUT);
                delay(500);                
                pinMode(DPIN_JOY_DOWN, INPUT);                 
                break;

            case 'm':
                pinMode(DPIN_JOY_FIRE, OUTPUT);
                delay(500);                
                pinMode(DPIN_JOY_FIRE, INPUT);                                 
                break;

            case 'f':
                disableFull ^= 1;                
                Serial.print("F");
                Serial.println(disableFull);
                break;

            default:
                break;
        }
    }
#endif  /* DEBUG_ENABLE */


    // -------------------- Input -----------------------------------------------------
    
    // Read analog joystick axes [0 .. ~1024] for a 100k pot
    joy1X = linearize(analogRead(APIN_JOY1X));
    joy1Y = linearize(analogRead(APIN_JOY1Y));

    joy2X = linearize(analogRead(APIN_JOY2X));
    joy2Y = linearize(analogRead(APIN_JOY2Y));
    

    // Detect calib push button 1
    if (joy1X > MAX_VALID_JOY_VALUE         // button pressed
        && joy1Y < MAX_VALID_JOY_VALUE)     // prevent false detection if no joystick is connected at all
    {
        joy1X = joy1X_old;
        calBtn1 = true;
    }

    // Detect calib push button 2
    calBtn2 = !digitalRead(DPIN_CALIB2);

    // Read joystick buttons. TODO: optimize by reading the port directly?
    joyBtn1 = !digitalRead(DPIN_JOYBTN1);
    joyBtn2 = !digitalRead(DPIN_JOYBTN2);
    joyBtn3 = !digitalRead(DPIN_JOYBTN3);
    joyBtn4 = !digitalRead(DPIN_JOYBTN4);
    

    // -------------------- Output of buttons -----------------------------------------

    switch (operatingMode)
    {
        case MODE_PADDLE:
        {
            // Trigger Paddle fire x on joystick button 1 or 3 and
            // fire y on joystick button 2 or 4
            bool fireX, fireY;
            static bool fireX_old = false;
            static bool fireY_old = false;
    
            fireX = joyBtn1 || joyBtn3;
            fireY = joyBtn2 || joyBtn4;
        
            if (fireX != fireX_old)
            {
                pinMode(DPIN_JOY_LEFT, fireX ? OUTPUT : INPUT);
            }
        
            if (fireY != fireY_old)
            {
                pinMode(DPIN_JOY_RIGHT, fireY ? OUTPUT : INPUT);
            }
        
            fireX_old = fireX;
            fireY_old = fireY;
            break;
        }    


        case MODE_JOY_WHEEL:
        case MODE_JOY_STICK:
        {
            // Fire <- joystick button 1 or 3
            // PotX <- joystick button 2 or 4 (high-aktive, Cheetah Annihilator compatible)
            bool fire, fire2;
            static bool fire_old = false;
            static bool fire2_old = false;
    
            fire = joyBtn1 || joyBtn3;
            fire2 = joyBtn2 || joyBtn4;
        
            if (fire != fire_old)
            {
                pinMode(DPIN_JOY_FIRE, fire ? OUTPUT : INPUT);
#if DEBUG_ENABLE
                if (debug == 2)                
                {
                    Serial.print('F');
                    Serial.println(fire);
                }
#endif                
            }
        
            if (fire2 != fire2_old)
            {
                if (fire2)
                {
                    digitalWrite(DPIN_PADDLE_FULL_X, HIGH);
                    digitalWrite(DPIN_PADDLE_PULSE_X, HIGH);
                }
                else
                {
                    digitalWrite(DPIN_PADDLE_FULL_X, LOW);
                    digitalWrite(DPIN_PADDLE_PULSE_X, LOW);
                }
            }
        
            fire_old = fire;
            fire2_old = fire2;
            break;
        }
    }


    /////////////////////////////////////////////////////////////////////////////////////
    // calculate median and moving average for smoothing of analogue values
    {
        // -------------- median filter ---------------------------
        medianBufX[medianIdx] = joy1X;
        medianBufY[medianIdx] = joy1Y;
        if (++medianIdx == MEDIAN_BUFLEN)
            medianIdx = 0;
    
        joy1X_median = median(medianBufX);
        joy1Y_median = median(medianBufY);


        // -------------- moving average filter -------------------
        // subtract oldest value from
        // and add latest value to sum        
        sumx -= bufx[idx];
        sumx += joy1X_median;
        sumy -= bufy[idx];
        sumy += joy1Y_median;

        // store new value in ring buffer
        bufx[idx] = joy1X_median;
        bufy[idx] = joy1Y_median;
        if (++idx == MA_BUFLEN)
            idx = 0;

        // Calculate mean
        joy1X_mean = sumx / MA_BUFLEN;
        joy1Y_mean = sumy / MA_BUFLEN;
    }

    
    /////////////////////////////////////////////////////////////////////////////////////
    // Main state machine
    switch (state)
    {       
        // ------------------------------------------------------------------------
        case NORMAL:
        {
            switch (operatingMode)
            {
                case MODE_PADDLE:
                {
                    bool xz, xz_old;
                    bool yz, yz_old;
                    
                    potX = joyToPaddle(joy1X_mean, axisX);      
                    tempX = paddleToOCR(potX); 
        
                    potY = joyToPaddle(joy1Y_mean, axisY);      
                    tempY = paddleToOCR(potY); 
                    
                    noInterrupts();
                    ocrX = tempX;
                    ocrY = tempY;
                    interrupts();
        
                    xz = (potX == 0);
                    yz = (potY == 0);
                    xz_old = (potX_old == 0);
                    yz_old = (potY_old == 0);
                    
#if DEBUG_ENABLE
                    if (disableFull)
                    {
                        xz_old = xz;
                        yz_old = yz;
                    }
#endif
            
        
                    if (xz != xz_old)
                    {
                        if (xz)
                        {
                            digitalWrite(DPIN_PADDLE_FULL_X, HIGH);
                            pinMode(DPIN_PADDLE_FULL_X, OUTPUT);
                        }
                        else
                        {
                            pinMode(DPIN_PADDLE_FULL_X, INPUT);
                        }
                    }
        
                    if (yz != yz_old)
                    {
                        if (yz)
                        {
                            digitalWrite(DPIN_PADDLE_FULL_Y, HIGH);
                            pinMode(DPIN_PADDLE_FULL_Y, OUTPUT);
                        }
                        else
                        {
                            pinMode(DPIN_PADDLE_FULL_Y, INPUT);
                        }
                    }

#if !DEBUG_ENABLE            
                    // Handle LED2
                    if (potX >= (axisX.pot.mid - LED_CENTER_THRESHOLD) && potX <= (axisX.pot.mid + LED_CENTER_THRESHOLD))
                        LED2_ON();
                    else
                        LED2_OFF();
#endif

                    break;
                }

                case MODE_JOY_WHEEL:
                case MODE_JOY_STICK:
                {                   
                    static int x = MID;
                    static int y = MID;
                    int xn, yn;
                    int threshLeft, threshRight, threshUp, threshDown;

                    // Set thresholds according to current state to realize a hysteresis
                    threshLeft = thresholdsLeft[x];
                    threshRight = thresholdsRight[x];

                    threshUp = thresholdsUp[y];
                    threshDown = thresholdsDown[y];

#if DEBUG_ENABLE
                    if (debug == 3)
                    {
                        Serial.print("x:");
                        Serial.print(x);
                        Serial.print(" thL:");
                        Serial.print(threshLeft);
                        Serial.print(" thR:");
                        Serial.print(threshRight);
                        Serial.print(" y:");
                        Serial.print(y);
                        Serial.print(" thU:");
                        Serial.print(threshUp);
                        Serial.print(" thD:");
                        Serial.println(threshDown);
                    }
#endif
                    
                    if (joy1X_mean < threshLeft)
                        xn = NEG;
                    else if (joy1X_mean > threshRight)
                        xn = POS;
                    else
                        xn = MID;

                    if (joy1Y_mean < threshUp)
                        yn = NEG;
                    else if (joy1Y_mean > threshDown)
                        yn = POS;
                    else
                        yn = MID;

                    if (xn != x)
                    {
#if DEBUG_ENABLE
                        if (debug == 3)
                        {
                            Serial.print("xn:");
                            Serial.println(xn);                           
                        }
#endif                    
                        
                        if (xn == MID)
                        {
                            pinMode(DPIN_JOY_LEFT, INPUT);
                            pinMode(DPIN_JOY_RIGHT, INPUT);                            
                        }
                        else if (xn == NEG)
                        {
                            pinMode(DPIN_JOY_LEFT, OUTPUT);
                            pinMode(DPIN_JOY_RIGHT, INPUT);
                        }
                        else if (xn == POS)
                        {
                            pinMode(DPIN_JOY_LEFT, INPUT);
                            pinMode(DPIN_JOY_RIGHT, OUTPUT);
                        }                        
                    }

                    if (yn != y)
                    {
#if DEBUG_ENABLE
                        if (debug == 3)
                        {
                            Serial.print("yn:");
                            Serial.println(yn);                           
                        }
#endif                    
                        if (yn == MID)
                        {
                            pinMode(DPIN_JOY_UP, INPUT);
                            pinMode(DPIN_JOY_DOWN, INPUT);                            
                        }
                        else if (yn == NEG)
                        {
                            pinMode(DPIN_JOY_UP, OUTPUT);
                            pinMode(DPIN_JOY_DOWN, INPUT);
                        }
                        else if (yn == POS)
                        {
                            pinMode(DPIN_JOY_UP, INPUT);
                            pinMode(DPIN_JOY_DOWN, OUTPUT);
                        }                        
                    }


#if !DEBUG_ENABLE            
                    // Handle LED2
                    if (x != MID || y != MID)
                        LED2_ON();
                    else
                        LED2_OFF();
#endif                        
                    
                    x = xn;
                    y = yn;
                    break;
                }

                
            } // switch (operatingMode)


#if DEBUG_ENABLE
            if (debug == 4)
            {
                dbgShowPinStates();               
            }
#endif


            // handle calib buttons
            if (calBtn1)
            {
                if (timeExpired(nextTime))
                {
                    state = CALIB_ENTER;
                }
            }
            else
            {
                nextTime = now + 2000;
            }            
            break;
        }

        // ------------------------------------------------------------------------
        case CALIB_ENTER:
        
            axisX.joy.max = 0;
            axisX.joy.mid = 511;
            axisX.joy.min = 1023;

            axisY.joy.max = 0;
            axisY.joy.mid = 511;
            axisY.joy.min = 1023;
        
            // Wait until calBtn1 released
            if (!calBtn1)
            {
                nextTime = now + 1000;       // wait some time before CALIB samples data to fill the moving average filter
                state = CALIB;                
            }
            break;

        
        // ------------------------------------------------------------------------
        case CALIB:
            if (!timeExpired(nextTime))
                break;

            if (joy1X_mean > axisX.joy.max)
                axisX.joy.max = joy1X_mean;
            else if (joy1X_mean < axisX.joy.min)
                axisX.joy.min = joy1X_mean;

            if (joy1Y_mean > axisY.joy.max)
                axisY.joy.max = joy1Y_mean;
            else if (joy1Y_mean < axisY.joy.min)
                axisY.joy.min = joy1Y_mean;

#if DEBUG_ENABLE
            Serial.print(joy1X);
            Serial.print("\t");
            Serial.print(joy1X_median);
            Serial.print("\t");
            Serial.print(joy1X_mean);
            Serial.print("\t");
//            Serial.print(axisX.joy.min);
//            Serial.print("\t");
//            Serial.print(axisX.joy.max);
//            Serial.print("\t");

            Serial.print(joy1Y);
            Serial.print("\t");
            Serial.print(joy1Y_median);
            Serial.print("\t");
            Serial.println(joy1Y_mean);
//            Serial.print("\t");
//            Serial.print(axisY.joy.min);
//            Serial.print("\t");
//            Serial.println(axisY.joy.max);
#endif

            // handle calib buttons
            if (calBtn1 || calBtn2 || joyBtn1 || joyBtn2)
            {               
                state = CENTER;     // calibrate center position
            }
            break;
        

        // ------------------------------------------------------------------------
        case CENTER:

//#if DEBUG_ENABLE 
//            Serial.print("Center:  ");
//            Serial.print(joy1X_mean);
//            Serial.print("\t");
//            Serial.print(joy1Y_mean);
//#endif

            axisX.joy.mid = joy1X_mean;
            axisY.joy.mid = joy1Y_mean;
            
            // Save calibration to EEPROM
            eepromWrite();
#if DEBUG_ENABLE
            printCalib();
#endif            

            // Handle if center position is near min or max
            if (fixCalibration())
            {
#if DEBUG_ENABLE
                Serial.println("Center pos near min or max => Fixed calibration:");
                printCalib();
#endif                
            }

            // Init thresholds for hysteresis
            initThresholds();

            state = NORMAL;
            nextTime = now + 2000;
            break;


#if DEBUG_ENABLE
        // ------------------------------------------------------------------------
        case OCR_OUT:
            noInterrupts();
            ocrX = ocrOutX;
            ocrY = ocrOutY;
            interrupts();            
            break;
#endif            
    } // switch (state)


    // handle LED1
    handleLed1(); 
      

#if DEBUG_ENABLE
    if (debug == 1)
    {
        if ((joy1X != joy1X_old) || (joy1Y != joy1Y_old) || calBtn1 || calBtn2)
        {
            Serial.print(calBtn1);
            Serial.print(calBtn2);
            Serial.print(" ");

            Serial.print(joyBtn1);
            Serial.print(joyBtn2);
            Serial.print(joyBtn3);
            Serial.print(joyBtn4);
            Serial.print(" ");        

            Serial.print(joy1X);
            Serial.print("\t");        
            Serial.print(joy1X_median);
            Serial.print("\t");        
            Serial.print(joy1X_mean);
            Serial.print("\t");        
            Serial.print(potX);
            Serial.print("\t");        

            Serial.print(joy1Y);
            Serial.print("\t");        
            Serial.print(joy1Y_median);
            Serial.print("\t");        
            Serial.print(joy1Y_mean);
            Serial.print("\t");        
            Serial.print(potY);
            Serial.print("\t");        

            Serial.print(joy2X);
            Serial.print("\t");        
            Serial.println(joy2Y);
        }
    }
#endif

    joy1X_old = joy1X;
    joy1Y_old = joy1Y;

    potX_old = potX;
    potY_old = potY;

    if (state == NORMAL)
        delayMicroseconds(100);
    else
        delayMicroseconds(1000);
}

/////////////////////////////////////////////////////////////////////
// Handle blinking of LED1
//
void handleLed1(void)
{
    static unsigned long nextLed1BlinkTime;
    static uint8_t blinkCounter = 0;
    const BlinkDef blinkDef[] = {
            {.blinkCount = 1, .onTime = 200, .offTime =   0, .pauseTime = 200},  // Calib state
            {.blinkCount = 1, .onTime = 100, .offTime =   0, .pauseTime = 1000}, // Paddle mode
            {.blinkCount = 2, .onTime = 100, .offTime = 200, .pauseTime = 1000}, // Joystick mode (Wheel optimized)
            {.blinkCount = 3, .onTime = 100, .offTime = 200, .pauseTime = 1000}, // Joystick mode (universal mode)
            {.blinkCount = 5, .onTime = 100, .offTime = 200, .pauseTime = 1000}, // Invalid mode
        };
    int blinkIdx;

    if (timeExpired(nextLed1BlinkTime))
    {
        if (state == NORMAL)
            blinkIdx = (operatingMode > MODE_NULL && operatingMode <= MODE_INVALID) ? operatingMode : MODE_INVALID;
        else
            blinkIdx = 0;


        if (led1State == LOW)
        {
            nextLed1BlinkTime += blinkDef[blinkIdx].onTime;
        }
        else
        {
            if (++blinkCounter < blinkDef[blinkIdx].blinkCount)
            {
                nextLed1BlinkTime += blinkDef[blinkIdx].offTime;
            }
            else
            {
                blinkCounter = 0;                
                nextLed1BlinkTime += blinkDef[blinkIdx].pauseTime;
            }
        }
        
        led1State ^= HIGH;
        digitalWrite(DPIN_LED1, led1State);
    }    
}

// The compare match interrupt triggers, when the SID discharges the paddle capacitor at the
// beginning of a new measurement cycle.
// The timer is set so that it generates a comapare match at the right time, charging the capacitor
// very quickly.
ISR(ANALOG_COMP_vect)
{
    uint16_t ti;
    

    // ignore IRQ that comes to early
    ti = TCNT1;
    if (ti < 4000 / F_CPU_DIV)
        return;
    
    
    // Arm Timer 1 (OC1A)
    TCCR1B = T1_STOP;       // Stop timer
    TCCR1A = T1_CLEAR;
    TCCR1C = (1 << FOC1A) | (1 << FOC1B);  // force compare match => clear output pins
    
    TCNT1 = 0;              // Clear timer
    OCR1A = ocrX;           // Set output compare match register
    OCR1B = ocrY;           // Set output compare match register
    
    TCCR1A = T1_SET;
    TCCR1B = T1_START;      // Start timer
}


int printOpMode(int mode)
{
    int ret = 0;
    
    Serial.print("Mode: ");
    if (mode == MODE_PADDLE)
        Serial.println("Paddle");
    else if (mode == MODE_JOY_WHEEL)
        Serial.println("Joystick (Wheel)");
    else if (mode == MODE_JOY_STICK)
        Serial.println("Joystick (Universal)");
    else
    {
        Serial.print(mode);
        Serial.println(" (Invalid)");
        ret = 1;
    }

    return ret;
}


#if DEBUG_ENABLE

void dbgShowPinStates(void)
{
    int fire, fire2;

    fire  = digitalReadPinMode(DPIN_JOY_FIRE);
    fire2 = digitalReadOutput(DPIN_PADDLE_FULL_X);

    Serial.print("F1:");
    Serial.print(fire);
    Serial.print(" F2:");
    Serial.print(fire2);

    Serial.print(" L:");
    Serial.print(digitalReadPinMode(DPIN_JOY_LEFT));
    Serial.print(" R:");
    Serial.print(digitalReadPinMode(DPIN_JOY_RIGHT));
    Serial.print(" U:");
    Serial.print(digitalReadPinMode(DPIN_JOY_UP));
    Serial.print(" D:");
    Serial.println(digitalReadPinMode(DPIN_JOY_DOWN));
}


int digitalReadPinMode(uint8_t pin)
{
    uint8_t bit = digitalPinToBitMask(pin);
    uint8_t port = digitalPinToPort(pin);

    if (port == NOT_A_PIN) return LOW;

    if (*portModeRegister(port) & bit) return HIGH;
    return LOW;
}



int digitalReadOutput(uint8_t pin)
{
    uint8_t bit = digitalPinToBitMask(pin);
    uint8_t port = digitalPinToPort(pin);

    if (port == NOT_A_PIN) return LOW;

    if (*portOutputRegister(port) & bit) return HIGH;
    return LOW;
}

#endif


// If joystick center position is near min or max, handle
// that depending on operatingMode.
bool fixCalibration(void)
{
    bool fixed = false;
    

    if (operatingMode == MODE_PADDLE)
    {
        int diff;

        // Make joy range a bit smaller, so that it can surely be reached

        diff = (axisX.joy.max - axisX.joy.min) / 64;
        axisX.joy.min += diff;
        axisX.joy.max -= diff;

        diff = (axisY.joy.max - axisY.joy.min) / 64;
        axisY.joy.min += diff;
        axisY.joy.max -= diff;

        
        // calulate a 'virtual' center in the middle of min and max if calibrated
        // center is near min or max.
        
        if ((axisX.joy.mid > (axisX.joy.max - MIN_CENTER_DISTANCE))
            || (axisX.joy.mid < (axisX.joy.min + MIN_CENTER_DISTANCE)))
        {
            axisX.joy.mid = (axisX.joy.max + axisX.joy.min) / 2;
            fixed = true;
        }

        if ((axisY.joy.mid > (axisY.joy.max - MIN_CENTER_DISTANCE))
            || (axisY.joy.mid < (axisY.joy.min + MIN_CENTER_DISTANCE)))
        {
            axisY.joy.mid = (axisY.joy.max + axisY.joy.min) / 2;
            fixed = true;
        }
    }
    else if (operatingMode == MODE_JOY_WHEEL || operatingMode == MODE_JOY_STICK)
    {        
        if (axisX.joy.mid > (axisX.joy.max - MIN_CENTER_DISTANCE))
        {
            axisX.joy.max += 1000;  // make unreachable
            fixed = true;
        }
        else if (axisX.joy.mid < (axisX.joy.min + MIN_CENTER_DISTANCE))
        {
            axisX.joy.min -= 1000;  // make unreachable
            fixed = true;            
        }

        if (axisY.joy.mid > (axisY.joy.max - MIN_CENTER_DISTANCE))
        {
            axisY.joy.max += 1000;  // make unreachable            
            fixed = true;
        }
        else if (axisY.joy.mid < (axisY.joy.min + MIN_CENTER_DISTANCE))
        {
            axisY.joy.min -= 1000;  // make unreachable            
            fixed = true;
        }
    }

    return fixed;
}
